/***************************************************************************************
 *      Copyright(c) 2014 ECONET Incorporation All rights reserved.
 *
 *      This is unpublished proprietary source code of ECONET Networks Incorporation
 *
 *      The copyright notice above does not evidence any actual or intended
 *      publication of such source code.
 ***************************************************************************************
 */

/*======================================================================================
 * MODULE NAME: spi
 * FILE NAME: spi_nand_flash_table.c
 * DATE: 2017/08/06
 * VERSION: 1.00
 * PURPOSE: To Provide SPI NAND Access interface.
 * NOTES:
 *======================================================================================
 */

#include <asm/arch-en7523/spi_nand_flash.h>

struct spi_nand_flash_ooblayout ooblayout_type1 = {
	.oobsize = 32, 
	.oobfree = {{0,8}, {16,8}, {32,8}, {48,8}}
};

struct spi_nand_flash_ooblayout ooblayout_type2 = {
	.oobsize = 50, 
	.oobfree = {{0,4}, {4,12}, {20,12}, {36,12}, {52,12} }
};

struct spi_nand_flash_ooblayout ooblayout_type3 = {
	.oobsize = 64, 
	.oobfree = {{0,64}}
};

struct spi_nand_flash_ooblayout ooblayout_type4 = {
	.oobsize = 128, 
	.oobfree = {{0,128}}
};

struct spi_nand_flash_ooblayout ooblayout_type5 = {
	.oobsize = 100, 
	.oobfree = {{0,4}, {4,12}, {20,12}, {36,12}, {52,12}, {68,12}, {84,12}, {100,12}, {116,12}}
};

struct spi_nand_flash_ooblayout ooblayout_type6 = {
	.oobsize = 33, 
	.oobfree = {{0,1}, {8,8}, {24,8}, {40,8}, {56,8}}
};

struct spi_nand_flash_ooblayout ooblayout_type7 = {
	.oobsize = 20, 
	.oobfree = {{0,4}, {4,4}, {20,4}, {36,4}, {52,4}}
};

struct spi_nand_flash_ooblayout ooblayout_type9 = {
	.oobsize = 64, 
	.oobfree = {{0,16}, {30,16}, {60,16}, {90,16}}
};

struct spi_nand_flash_ooblayout ooblayout_type10 = {
	.oobsize = 72, 
	.oobfree = {{0,18}, {32,18}, {64,18}, {96,18}}
};

struct spi_nand_flash_ooblayout ooblayout_type11 = {
	.oobsize = 144, 
	.oobfree = {{0,18}, {32,18}, {64,18}, {96,18}, {128,18}, {160,18}, {192,18}, {224,18}}
};

struct spi_nand_flash_ooblayout ooblayout_type12 = {
	.oobsize = 144, 
	.oobfree = {{0,144}}
};

struct spi_nand_flash_ooblayout ooblayout_type13 = {
	.oobsize = 36, 
	.oobfree = {{0,4}, {32,32}}
};

struct spi_nand_flash_ooblayout ooblayout_type15 = {
	.oobsize = 36, 
	.oobfree = {{0,4}, {4,8}, {20,8}, {36,8}, {52,8} }
};

struct spi_nand_flash_ooblayout ooblayout_type18 = {
	.oobsize = 96, 
	.oobfree = {{0,24}, {32,24},  {64,24}, {96,24}}
};

struct spi_nand_flash_ooblayout ooblayout_type19 = {
	.oobsize = 44, 
	.oobfree = {{0,4}, {8,40}}
};


/* ===== SPI NAND Controller ECC ooblayout ===== */
struct spi_nand_flash_ooblayout ooblayout_spi_controller_ecc;

/*****************************[ Notice]******************************/
/* If new spi nand chip have page size more than 4KB,  or oob size more than 256 bytes,  than*/
/* it will need to adjust the #define of _SPI_NAND_PAGE_SIZE and _SPI_NAND_OOB_SIZE */
/*****************************[ Notice]******************************/

const struct SPI_NAND_FLASH_INFO_T spi_nand_flash_tables[] = {
	/* ===== Flash ECC ===== */
	
	/* ===== GigaDevice ===== */
	{
		mfr_id:						_SPI_NAND_MANUFACTURER_ID_GIGADEVICE,
		dev_id:						_SPI_NAND_DEVICE_ID_GD5F1GQ4UBYIG,
		ptr_name:					"_SPI_NAND_DEVICE_ID_GD5F1GQ4UB",
		device_size:				_SPI_NAND_CHIP_SIZE_1GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_128BYTE,
		erase_size:					_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode:					SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout : 			&ooblayout_type2,
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_LOAD_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},
	{
		mfr_id:						_SPI_NAND_MANUFACTURER_ID_GIGADEVICE,
		dev_id:						_SPI_NAND_DEVICE_ID_GD5F2GQ4UBYIG,
		ptr_name:					"_SPI_NAND_DEVICE_ID_GD5F2GQ4UB",
		device_size:				_SPI_NAND_CHIP_SIZE_2GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_128BYTE,
		erase_size:					_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode:					SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout : 			&ooblayout_type2,
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_LOAD_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},

	{
		mfr_id:						_SPI_NAND_MANUFACTURER_ID_GIGADEVICE,
		dev_id:						_SPI_NAND_DEVICE_ID_GD5F2GQ4UE9IS,
		ptr_name:					"_SPI_NAND_DEVICE_ID_GD5F2GQ4UE9IS",
		device_size:				_SPI_NAND_CHIP_SIZE_2GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_128BYTE,
		erase_size:					_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode:					SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout : 			&ooblayout_type2,
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_LOAD_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},
	{
		mfr_id:						_SPI_NAND_MANUFACTURER_ID_GIGADEVICE,
		dev_id:						_SPI_NAND_DEVICE_ID_GD5F4GQ4UBYIG,
		ptr_name:					"_SPI_NAND_DEVICE_ID_GD5F4GQ4UB",
		device_size:				_SPI_NAND_CHIP_SIZE_4GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_4KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_256BYTE,
		erase_size:					_SPI_NAND_BLOCK_SIZE_256KBYTE,
		dummy_mode:					SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout : 			&ooblayout_type5,
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_LOAD_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},	
	/* ===== ESMT ===== */
	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_ESMT,
		dev_id: 					_SPI_NAND_DEVICE_ID_F50L1G41A0,
		ptr_name:					"_SPI_NAND_DEVICE_ID_F50L1G",
		device_size:				_SPI_NAND_CHIP_SIZE_1GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_64BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout : 			&ooblayout_type6,
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x0, 0x0},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},	

	{
		mfr_id:						_SPI_NAND_MANUFACTURER_ID_ESMT,
		dev_id:						_SPI_NAND_DEVICE_ID_F50L1G41LB,
		ptr_name:					"_SPI_NAND_DEVICE_ID_F50L1G41LB",
		device_size:				_SPI_NAND_CHIP_SIZE_1GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_64BYTE,
		erase_size:					_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode:					SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout : 			&ooblayout_type7,
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x7C, 0x0},
		quad_en:					{0x0, 0x0},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},

	{
		mfr_id:						_SPI_NAND_MANUFACTURER_ID_ESMT,
		dev_id:						_SPI_NAND_DEVICE_ID_F50L2G41LB,
		ptr_name:					"_SPI_NAND_DEVICE_ID_F50L2G41LB",
		device_size:				_SPI_NAND_CHIP_SIZE_2GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_64BYTE,
		erase_size:					_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode:					SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout : 			&ooblayout_type7,
		feature:					SPI_NAND_FLASH_DIE_SELECT_1_HAVE,
		die_num:					2,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x7C, 0x0},
		quad_en:					{0x0, 0x0},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},

	/* ===== Winbond ===== */
	{
		mfr_id:						_SPI_NAND_MANUFACTURER_ID_WINBOND,
		dev_id:						_SPI_NAND_DEVICE_ID_W25N01GV,
		ptr_name:					"_SPI_NAND_DEVICE_ID_W25N01G",
		device_size:				_SPI_NAND_CHIP_SIZE_1GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_64BYTE,
		erase_size:					_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode:					SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout : 			&ooblayout_type7,
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0xFB, 0x0},
		quad_en:					{0x0, 0x0},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},

	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_WINBOND,
		dev_id: 					_SPI_NAND_DEVICE_ID_W25M02GV,
		ptr_name:					"_SPI_NAND_DEVICE_ID_W25M02G",
		device_size:				_SPI_NAND_CHIP_SIZE_2GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_64BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout : 			&ooblayout_type7,
		feature:					SPI_NAND_FLASH_DIE_SELECT_1_HAVE,
		die_num:					2,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0xFB, 0x0},
		quad_en:					{0x0, 0x0},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},		

	/* ===== MXIC ===== */
	{
		mfr_id:						_SPI_NAND_MANUFACTURER_ID_MXIC,
		dev_id:						_SPI_NAND_DEVICE_ID_MXIC35LF1GE4AB,
		ptr_name:					"_SPI_NAND_DEVICE_ID_MXIC35LF1G",
		device_size:				_SPI_NAND_CHIP_SIZE_1GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_64BYTE,
		erase_size:					_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode:					SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout : 			&ooblayout_type7,
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},

	{
		mfr_id:						_SPI_NAND_MANUFACTURER_ID_MXIC,
		dev_id:						_SPI_NAND_DEVICE_ID_MXIC35LF2GE4AB,
		ptr_name:					"_SPI_NAND_DEVICE_ID_MXIC35LF2G",
		device_size:				_SPI_NAND_CHIP_SIZE_2GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_64BYTE,
		erase_size:					_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode:					SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout : 			&ooblayout_type2,
		feature:					SPI_NAND_FLASH_PLANE_SELECT_HAVE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},

	/* ===== Zentel ===== */
	/* ===== Etron ===== */
	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_ETRON,
		dev_id: 					_SPI_NAND_DEVICE_ID_EM73C044SNB,
		ptr_name:					"_SPI_NAND_DEVICE_ID_EM73C044SNB",
		device_size:				_SPI_NAND_CHIP_SIZE_1GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_128BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout : 			&ooblayout_type9,		
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},

	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_ETRON,
		dev_id: 					_SPI_NAND_DEVICE_ID_EM73C044SND,
		ptr_name:					"_SPI_NAND_DEVICE_ID_EM73C044SND",
		device_size:				_SPI_NAND_CHIP_SIZE_1GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_64BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout : 			&ooblayout_type1,		
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},

	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_ETRON,
		dev_id: 					_SPI_NAND_DEVICE_ID_EM73C044SNF,
		ptr_name:					"_SPI_NAND_DEVICE_ID_EM73C044SNF",
		device_size:				_SPI_NAND_CHIP_SIZE_1GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_128BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout : 			&ooblayout_type10,		
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},

	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_ETRON,
		dev_id: 					_SPI_NAND_DEVICE_ID_EM73C044VCA,
		ptr_name:					"_SPI_NAND_DEVICE_ID_EM73C044VCA",
		device_size:				_SPI_NAND_CHIP_SIZE_1GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_64BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout : 			&ooblayout_type1,		
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},

	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_ETRON,
		dev_id: 					_SPI_NAND_DEVICE_ID_EM73C044VCD,
		ptr_name:					"_SPI_NAND_DEVICE_ID_EM73C044VCD",
		device_size:				_SPI_NAND_CHIP_SIZE_1GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_64BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout : 			&ooblayout_type1,		
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},

	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_ETRON,
		dev_id: 					_SPI_NAND_DEVICE_ID_EM73D044SNA,
		ptr_name:					"_SPI_NAND_DEVICE_ID_EM73D044SNA",
		device_size:				_SPI_NAND_CHIP_SIZE_2GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_128BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout :			&ooblayout_type10,		
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},

	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_ETRON,
		dev_id: 					_SPI_NAND_DEVICE_ID_EM73D044SNC,
		ptr_name:					"_SPI_NAND_DEVICE_ID_EM73D044SNC",
		device_size:				_SPI_NAND_CHIP_SIZE_2GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_128BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout :			&ooblayout_type9,		
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},

	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_ETRON,
		dev_id: 					_SPI_NAND_DEVICE_ID_EM73D044SND,
		ptr_name:					"_SPI_NAND_DEVICE_ID_EM73D044SND",
		device_size:				_SPI_NAND_CHIP_SIZE_2GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_64BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout :			&ooblayout_type1,
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},

	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_ETRON,
		dev_id: 					_SPI_NAND_DEVICE_ID_EM73D044SNF,
		ptr_name:					"_SPI_NAND_DEVICE_ID_EM73D044SNF",
		device_size:				_SPI_NAND_CHIP_SIZE_2GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_128BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout :			&ooblayout_type10,		
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},

	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_ETRON,
		dev_id: 					_SPI_NAND_DEVICE_ID_EM73D044VCA,
		ptr_name:					"_SPI_NAND_DEVICE_ID_EM73D044VCA",
		device_size:				_SPI_NAND_CHIP_SIZE_2GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_128BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout :			&ooblayout_type18,		
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},

	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_ETRON,
		dev_id: 					_SPI_NAND_DEVICE_ID_EM73D044VCB,
		ptr_name:					"_SPI_NAND_DEVICE_ID_EM73D044VCB",
		device_size:				_SPI_NAND_CHIP_SIZE_2GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_64BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout :			&ooblayout_type1,		
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},

	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_ETRON,
		dev_id: 					_SPI_NAND_DEVICE_ID_EM73D044VCD,
		ptr_name:					"_SPI_NAND_DEVICE_ID_EM73D044VCD",
		device_size:				_SPI_NAND_CHIP_SIZE_2GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_128BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout :			&ooblayout_type10,		
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},

	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_ETRON,
		dev_id: 					_SPI_NAND_DEVICE_ID_EM73D044VCG,
		ptr_name:					"_SPI_NAND_DEVICE_ID_EM73D044VCG",
		device_size:				_SPI_NAND_CHIP_SIZE_2GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_64BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout :			&ooblayout_type1,		
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},

	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_ETRON,
		dev_id: 					_SPI_NAND_DEVICE_ID_EM73D044VCH,
		ptr_name:					"_SPI_NAND_DEVICE_ID_EM73D044VCH",
		device_size:				_SPI_NAND_CHIP_SIZE_2GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_64BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout :			&ooblayout_type1,		
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},

	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_ETRON,
		dev_id: 					_SPI_NAND_DEVICE_ID_EM73E044SNA,
		ptr_name:					"_SPI_NAND_DEVICE_ID_EM73E044SNA",
		device_size:				_SPI_NAND_CHIP_SIZE_4GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_4KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_256BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_256KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout :			&ooblayout_type11,		
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},

	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_ETRON,
		dev_id: 					_SPI_NAND_DEVICE_ID_EM73F044SNA,
		ptr_name:					"_SPI_NAND_DEVICE_ID_EM73F044SNA",
		device_size:				_SPI_NAND_CHIP_SIZE_8GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_4KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_256BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_256KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout :			&ooblayout_type12,		
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},

	/* ===== Toshiba ===== */
	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_TOSHIBA,
		dev_id: 					_SPI_NAND_DEVICE_ID_TC58CVG0S3H,
		ptr_name:					"_SPI_NAND_DEVICE_ID_TC58CVG0S3H",
		device_size:				_SPI_NAND_CHIP_SIZE_1GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_64BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout : 			&ooblayout_type3,		
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x0, 0x0},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},		

	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_TOSHIBA,
		dev_id: 					_SPI_NAND_DEVICE_ID_TC58CVG1S3H,
		ptr_name:					"_SPI_NAND_DEVICE_ID_TC58CVG1S3H",
		device_size:				_SPI_NAND_CHIP_SIZE_2GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_64BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout : 			&ooblayout_type3,		
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x0, 0x0},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	}, 

	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_TOSHIBA,
		dev_id: 					_SPI_NAND_DEVICE_ID_TC58CVG2S0H,
		ptr_name:					"_SPI_NAND_DEVICE_ID_TC58CVG2S0H",
		device_size:				_SPI_NAND_CHIP_SIZE_4GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_4KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_128BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_256KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout : 			&ooblayout_type4,		
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x0, 0x0},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	}, 

	/* ===== Micron ===== */
	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_MICRON,
		dev_id: 					_SPI_NAND_DEVICE_ID_MT29F1G01,
		ptr_name:					"_SPI_NAND_DEVICE_ID_MT29F1G01",
		device_size:				_SPI_NAND_CHIP_SIZE_1GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_128BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout :			&ooblayout_type13, 	
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x70, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x7C, 0x0},
		quad_en:					{0x0, 0x0},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},		

	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_MICRON,
		dev_id: 					_SPI_NAND_DEVICE_ID_MT29F2G01,
		ptr_name:					"_SPI_NAND_DEVICE_ID_MT29F2G01",
		device_size:				_SPI_NAND_CHIP_SIZE_2GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_128BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout :			&ooblayout_type13, 	
		feature:					SPI_NAND_FLASH_PLANE_SELECT_HAVE,
		die_num:					1,
		ecc_fail_check_info:		{0x70, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x7C, 0x0},
		quad_en:					{0x0, 0x0},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},

	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_MICRON,
		dev_id: 					_SPI_NAND_DEVICE_ID_MT29F4G01,
		ptr_name:					"_SPI_NAND_DEVICE_ID_MT29F4G01",
		device_size:				_SPI_NAND_CHIP_SIZE_4GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_128BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout :			&ooblayout_type13, 	
		feature:					SPI_NAND_FLASH_PLANE_SELECT_HAVE | SPI_NAND_FLASH_DIE_SELECT_2_HAVE,
		die_num:					2,
		ecc_fail_check_info:		{0x70, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x7C, 0x0},
		quad_en:					{0x0, 0x0},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},
	/* ===== ATO ===== */
	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_ATO,
		dev_id: 					_SPI_NAND_DEVICE_ID_ATO25D1GA,
		ptr_name:					"_SPI_NAND_DEVICE_ID_ATO25D1GA",
		device_size:				_SPI_NAND_CHIP_SIZE_1GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_64BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_SINGLE,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout :			&ooblayout_type3, 	
		feature:					SPI_NAND_FLASH_NO_ECC_STATUS_HAVE,
		die_num:					1,
		ecc_fail_check_info:		{0x0, 0x0},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x00, 0x00},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},

	/* ===== FM ===== */
	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_FM,
		dev_id: 					_SPI_NAND_DEVICE_ID_FM25S01,
		ptr_name:					"_SPI_NAND_DEVICE_ID_FM25S01",
		device_size:				_SPI_NAND_CHIP_SIZE_1GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_128BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout :			&ooblayout_type3, 	
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x78, 0x0},
		quad_en:					{0x00, 0x00},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},
	
	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_FM,
		dev_id: 					_SPI_NAND_DEVICE_ID_FM25G01B,
		ptr_name:					"_SPI_NAND_DEVICE_ID_FM25G01B",
		device_size:				_SPI_NAND_CHIP_SIZE_1GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_128BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout :			&ooblayout_type3, 	
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x70, 0x70},
		write_en_type:				SPI_NAND_FLASH_WRITE_LOAD_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_ECC, 0x10, 0x10},
	},

	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_FM,
		dev_id: 					_SPI_NAND_DEVICE_ID_FM25G02B,
		ptr_name:					"_SPI_NAND_DEVICE_ID_FM25G02B",
		device_size:				_SPI_NAND_CHIP_SIZE_2GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_128BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout :			&ooblayout_type3, 	
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x70, 0x70},
		write_en_type:				SPI_NAND_FLASH_WRITE_LOAD_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_ECC, 0x10, 0x10},
	},

	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_FM,
		dev_id: 					_SPI_NAND_DEVICE_ID_FM25G02,
		ptr_name:					"_SPI_NAND_DEVICE_ID_FM25G02",
		device_size:				_SPI_NAND_CHIP_SIZE_2GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_64BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout :			&ooblayout_type1, 	
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x70, 0x70},
		write_en_type:				SPI_NAND_FLASH_WRITE_LOAD_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_ECC, 0x10, 0x10},
	},

	/* ===== XTX ===== */
	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_XTX,
		dev_id: 					_SPI_NAND_DEVICE_ID_XT26G02B,
		ptr_name:					"_SPI_NAND_DEVICE_ID_XT26G02B",
		device_size:				_SPI_NAND_CHIP_SIZE_2GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_64BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout :			&ooblayout_type1, 	
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x70, 0x70},
		write_en_type:				SPI_NAND_FLASH_WRITE_LOAD_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_ECC, 0x10, 0x10},
	},

	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_XTX,
		dev_id: 					_SPI_NAND_DEVICE_ID_XT26G01A,
		ptr_name:					"_SPI_NAND_DEVICE_ID_XT26G01A",
		device_size:				_SPI_NAND_CHIP_SIZE_1GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_64BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout :			&ooblayout_type19, 	
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x3C, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_LOAD_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},

	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_XTX,
		dev_id: 					_SPI_NAND_DEVICE_ID_XT26G02A,
		ptr_name:					"_SPI_NAND_DEVICE_ID_XT26G02A",
		device_size:				_SPI_NAND_CHIP_SIZE_2GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_64BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout :			&ooblayout_type19, 	
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_LOAD_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},


	/* ===== MIRA ===== */
	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_MIRA,
		dev_id: 					_SPI_NAND_DEVICE_ID_PSU1GS20BN,
		ptr_name:					"_SPI_NAND_DEVICE_ID_PSU1GS20BN",
		device_size:				_SPI_NAND_CHIP_SIZE_1GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_64BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout : 			&ooblayout_type6,
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x0, 0x0},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},

	/* ===== BIWIN ===== */
	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_BIWIN,
		dev_id: 					_SPI_NAND_DEVICE_ID_BWJX08U,
		ptr_name:					"_SPI_NAND_DEVICE_ID_BWJX08U",
		device_size:				_SPI_NAND_CHIP_SIZE_2GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_64BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout :			&ooblayout_type15, 	
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},

	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_BIWIN,
		dev_id: 					_SPI_NAND_DEVICE_ID_BWET08U,
		ptr_name:					"_SPI_NAND_DEVICE_ID_BWET08U",
		device_size:				_SPI_NAND_CHIP_SIZE_2GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_128BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout :			&ooblayout_type10, 	
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x30, 0x20},
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x10, 0x10},
	},

	/* ===== FORESEE ===== */
	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_FORESEE,
		dev_id: 					_SPI_NAND_DEVICE_ID_FS35ND01GD1F1,
		ptr_name:					"_SPI_NAND_DEVICE_ID_FS35ND01GD1F1",
		device_size:				_SPI_NAND_CHIP_SIZE_1GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_64BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout :			&ooblayout_type1, 	
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x70, 0x70},
		write_en_type:				SPI_NAND_FLASH_WRITE_LOAD_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_ECC, 0x10, 0x10},
	},
	
	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_FORESEE,
		dev_id: 					_SPI_NAND_DEVICE_ID_FS35ND02GS2F1,
		ptr_name:					"_SPI_NAND_DEVICE_ID_FS35ND02GS2F1",
		device_size:				_SPI_NAND_CHIP_SIZE_2GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_128BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout :			&ooblayout_type1, 	
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x70, 0x70},
		write_en_type:				SPI_NAND_FLASH_WRITE_LOAD_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_ECC, 0x10, 0x10},
	},

	{
		mfr_id: 					_SPI_NAND_MANUFACTURER_ID_FORESEE,
		dev_id: 					_SPI_NAND_DEVICE_ID_FS35ND02GD1F1,
		ptr_name:					"_SPI_NAND_DEVICE_ID_FS35ND02GD1F1",
		device_size:				_SPI_NAND_CHIP_SIZE_2GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_128BYTE,
		erase_size: 				_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode: 				SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout :			&ooblayout_type1, 	
		feature:					SPI_NAND_FLASH_FEATURE_NONE,
		die_num:					1,
		ecc_fail_check_info:		{0x70, 0x70},
		write_en_type:				SPI_NAND_FLASH_WRITE_LOAD_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_ECC, 0x10, 0x10},
	},

	/* ===== SPI controller ECC ===== */
		
	/* ===== MXIC ===== */
	{
		mfr_id:						_SPI_NAND_MANUFACTURER_ID_MXIC,
		dev_id:						_SPI_NAND_DEVICE_ID_MXIC35LF2G14AC,
		ptr_name:					"_SPI_NAND_DEVICE_ID_MXIC35LF2G14AC",
		device_size:				_SPI_NAND_CHIP_SIZE_2GBIT,
		page_size:					_SPI_NAND_PAGE_SIZE_2KBYTE,
		oob_size:					_SPI_NAND_OOB_SIZE_64BYTE,
		erase_size:					_SPI_NAND_BLOCK_SIZE_128KBYTE,
		dummy_mode:					SPI_NAND_FLASH_READ_DUMMY_BYTE_APPEND,
		read_mode:					SPI_NAND_FLASH_READ_SPEED_MODE_DUAL,
		write_mode:					SPI_NAND_FLASH_WRITE_SPEED_MODE_SINGLE,
		oob_free_layout : 			&ooblayout_spi_controller_ecc,
		feature:					SPI_NAND_FLASH_PLANE_SELECT_HAVE,
		die_num:					1,
		write_en_type:				SPI_NAND_FLASH_WRITE_EN_FIRST,
		unlock_block_info:			{0x38, 0x0},
		quad_en:					{0x01, 0x01},
		ecc_en:						{_SPI_NAND_ADDR_FEATURE, 0x00, 0x00},
	},
};

SPI_NAND_FLASH_RTN_T scan_spi_nand_table(struct SPI_NAND_FLASH_INFO_T *ptr_rtn_device_t)
{
	u32 i = 0;
	SPI_NAND_FLASH_RTN_T rtn_status = SPI_NAND_FLASH_RTN_PROBE_ERROR;
	
	for (i = 0; i < (sizeof(spi_nand_flash_tables) / sizeof(spi_nand_flash_tables[0])) ; i++) 
	{
		if ( ( (ptr_rtn_device_t->mfr_id) == spi_nand_flash_tables[i].mfr_id) &&
		     ( (ptr_rtn_device_t->dev_id) == spi_nand_flash_tables[i].dev_id)  ) 
		{
			ptr_rtn_device_t->device_size 			= spi_nand_flash_tables[i].device_size;
			ptr_rtn_device_t->erase_size  			= spi_nand_flash_tables[i].erase_size;
			ptr_rtn_device_t->page_size	  			= spi_nand_flash_tables[i].page_size;
			ptr_rtn_device_t->oob_size	  			= spi_nand_flash_tables[i].oob_size;
			ptr_rtn_device_t->dummy_mode  			= spi_nand_flash_tables[i].dummy_mode;
			ptr_rtn_device_t->read_mode	  			= spi_nand_flash_tables[i].read_mode;
			ptr_rtn_device_t->write_mode			= spi_nand_flash_tables[i].write_mode;
			memcpy( &(ptr_rtn_device_t->ptr_name) , &(spi_nand_flash_tables[i].ptr_name), sizeof(ptr_rtn_device_t->ptr_name));
			memcpy( &(ptr_rtn_device_t->oob_free_layout) , &(spi_nand_flash_tables[i].oob_free_layout), sizeof(ptr_rtn_device_t->oob_free_layout));
			ptr_rtn_device_t->feature = spi_nand_flash_tables[i].feature;
			ptr_rtn_device_t->die_num = spi_nand_flash_tables[i].die_num;
			memcpy( &(ptr_rtn_device_t->ecc_fail_check_info) , &(spi_nand_flash_tables[i].ecc_fail_check_info), sizeof(struct SPI_NAND_ECC_FAIL_CHECK_INFO_T));
			ptr_rtn_device_t->write_en_type = spi_nand_flash_tables[i].write_en_type;
			memcpy( &(ptr_rtn_device_t->unlock_block_info) , &(spi_nand_flash_tables[i].unlock_block_info), sizeof(struct SPI_NAND_UNLOCK_BLOCK_INFO_T));
			memcpy( &(ptr_rtn_device_t->quad_en) , &(spi_nand_flash_tables[i].quad_en), sizeof(struct SPI_NAND_QUAD_EN_INFO_T));
			memcpy( &(ptr_rtn_device_t->ecc_en) , &(spi_nand_flash_tables[i].ecc_en), sizeof(struct SPI_NAND_ECC_EN_INFO_T));

			rtn_status = SPI_NAND_FLASH_RTN_NO_ERROR;
			break;
		}
	}

	return rtn_status;
}

/* End of [spi_nand_flash_table.c] package */
