#include <linux/version.h>
#include <asm/tc3162.h>

#include "skbuff.h"
#include "eth.h"

#include <linux/types.h>

#include "ecnt_eth.h"

void pause(int ms)
{
	unsigned int i;
	for (i = 0; i < ms * 100000; i++)
		;
}

#define TCPHY_4PORT 1
#define TC2104MC_SUPPORT 1
#define TCPHY_SUPPORT 1

#define DO_PER_PORT 1
#define DO_4_PORT 0
#define DO_5_PORT 2

#define SWIC_TC2206 7
#define tcPhyVer_2031 0 // 9400, LEH
#define tcPhyVer_2101mb  1 // 9401, LEM
#define tcPhyVer_2104mc  2 // 9402, tc2206
#define tcPhyVer_2104sd  3 // 9403, tc2104sd
#define tcPhyVer_2101me  4 // 9404, 62UE
#define tcPhyVer_2102me  5 // 9405, tc3182
#define tcPhyVer_2104me  6 // 9406, tc2206F
#define tcPhyVer_2101mf  7 // 9407, FPGA(MF)
#define tcPhyVer_2105mj  11 // 940b, RT63365
#define TCPHY_PORTNUM 4
#define mtPhyVer_7530    13 //MT7530
#define tcPhyVer_mt7510FE 15 //MT7510,FE(Fast Ethernet)
#define mtPhyVer_7510Ge  16 //MT7510Ge      

#define tcPhyVer_7512Fe		18	// 9412, EN7512Fe, need to check R15(0xf)[13]=0: no Giga capability
#define tcPhyVer_7512Ge		19	// 9451, EN7512Ge
#define tcPhyVer_7522Fe		20	// 9414
#define tcPhyVer_7522Ge		21	// 9471	


#define TCPHYDISP1 if(tcPhy_disp_level>=1) printf
#define TCPHYDISP4 if(tcPhy_disp_level>=4) printf
#define TCPHYDISP5 if(tcPhy_disp_level>=5) printf

#define bool uint8 
#define ENABLE 1
#define DISABLE 0
#define tc2104mc_G0R22_Tx10AmpSave_ON  0x0e64
#define  tc2104mc_G0R22_Tx10AmpSave_OFF  0x066F

#define printf prom_printf

#define tcMiiStationRead(x, y)   miiStationRead(x, y)
#define tcMiiStationWrite(x, y, z) miiStationWrite(x, y, z)

// for tcXXXX_link_state
#define ST_LINK_DOWN 0
static uint32 eco_rev = 0x00;
extern macAdapter_t *mac_p;


static phyDeviceList_t phyDeviceList[] = {
	{ 0x0013, "LXT972" },
	{ 0x0022, "AC101L" },
	{ 0x0243, "IP101" },
	{ 0x8201, "RTL8201" },
	{ 0x03a2, "TC2031" }
};
static uint8	ephy_addr_base = 0;		// replace  "mac_p->enetPhyAddr"


#define MODEL_V0	(1<<0)	/*TC2206A1*/
#define MODEL_V1	(1<<1)	/*TC2206B2*/
#define MODEL_V2	(1<<2)	/*TC2206F*/

typedef struct swic_reg{
	uint16 model_id;
	uint16 phyaddr;
	uint16 regaddr;
	uint16 value;
}swic_reg_t;

swic_reg_t tc2206_swic_reg[]=
{
	{MODEL_V0|MODEL_V1, 22, 0, 0x6f9},/*P4MECR,Enable RGMII Mode*/
	{MODEL_V0|MODEL_V1, 22, 1, 0x7f8},/*P5MECR*/
	{MODEL_V0|MODEL_V1, 21, 5, 0x68c},/*HIGMPSCR*/
	{MODEL_V0|MODEL_V1, 20, 13, 0x20},/*ATCR*/
	{MODEL_V0|MODEL_V1, 21, 17, 0x10d},/*SCR1*/
	{MODEL_V0|MODEL_V1, 21, 18, 0x7fc},/*P0MCR*/
	{MODEL_V0|MODEL_V1, 21, 19, 0x7fc},/*P1MCR*/
	{MODEL_V0|MODEL_V1, 21, 20, 0x7fc},/*P2MCR*/
	{MODEL_V0|MODEL_V1, 21, 21, 0x7fc},/*P3MCR*/
	{MODEL_V0|MODEL_V1, 21, 22, 0x0ffc},/*P4MCR*/
	{MODEL_V0|MODEL_V1|MODEL_V2, 21, 23, 0x5ffc},/*P5MCR,Let CPU Port (Port 5) fixed on Link UP & 100 Full-Duplex Speed*/
	{MODEL_V0|MODEL_V1, 20, 2, 0xf20f},/*P0CR*/
	{MODEL_V0|MODEL_V1, 20, 3, 0xf20f},/*P1CR*/
	{MODEL_V0|MODEL_V1, 20, 4, 0xf20f},/*P2CR*/
	{MODEL_V0|MODEL_V1, 20, 5, 0xf20f},/*P3CR*/
	{MODEL_V0|MODEL_V1, 20, 6, 0xf20f},/*P4CR*/
	{MODEL_V0|MODEL_V1, 20, 7, 0xf20f},/*P5CR*/
	{MODEL_V0|MODEL_V1, 21, 31, 0x40},/*MCMBTCR*/
	{MODEL_V0|MODEL_V1|MODEL_V2, 24, 22, 0x4000},/*Enable the secondary L2 mac table to workaround thatSA=0.0.0.0 will be used entry 2049 entry*/
	{MODEL_V0|MODEL_V1, 30, 8, 0x7777},/*RXCDC, must to be set up*/
	{MODEL_V0|MODEL_V1, 30, 13, 0x0},/*WDTCR,Disable the watchdog timer*/
	{MODEL_V0|MODEL_V1|MODEL_V2, 21, 4, 0xd},/*MLDPCR, MLD(IPv6 packet) forward to port_dft_mcas*/
	{MODEL_V0|MODEL_V1|MODEL_V2, 21, 16, 0x10},/*SCR0,Disable illegal DA/SA check operation*/
	{0, 0, 0x0}
};

/*Those registers is used to check the switch register is reset to default value.*/
swic_reg_t esd_chk_reg[]=
{
	{MODEL_V0|MODEL_V1, 21, 22, 0x0},/*P4MCR*/
	{MODEL_V0|MODEL_V1|MODEL_V2, 21, 23, 0x0},/*P5MCR,Let CPU Port (Port 5) fixed on Link UP & 100 Full-Duplex Speed*/
	{MODEL_V0|MODEL_V1, 30, 8, 0x0},/*RXCDC, must to be set up*/
	{MODEL_V0|MODEL_V1, 30, 13, 0x0},/*WDTCR,Disable the watchdog timer*/
	{0,0, 0, 0x0}	
};

static uint8 tcPhyFlag = 0;
static uint8 tcPhyVer = 99; // default: unknown PhyVer
static uint8 tcPhyPortNum = 1; // set value in tcPhyVerLookUp
static uint8 swicVendor = 0;
static uint8 esd_sw_patch=1;//default is enable ESD software patch
static uint8 sw_patch_flag = 1;
uint8 tcPhy_disp_level = 0; // default level 0
static uint8 current_idx = 0; // default 0, Do NOT change 
static uint8 sw_FixPdWakeup_flag = 1;/*Default is enable*/
static uint8 tcphy_link_state[TCPHY_PORTNUM];
static uint8 cfg_Tx10AmpSave_flag = 1; // default ENABLE
static uint8 sw_ErrOverMonitor_flag = 1; // default enable
static uint8 tc2104mc_FixNonAnPairSwap_flag = 1;// default: enable

// variables for mii registers
typedef struct {
	uint8 main_reset; // 15
	uint8 force_speed; // 13
	uint8 autoneg_enable; // 12
	uint8 powerdown; // 11
	uint8 force_duplex; // 8
} tcphy_mr0_reg_t;

typedef struct {
	uint16 value; // 15:0
	bool link_status; // 2
	bool link_status_prev;
} tcphy_mr1_reg_t;

typedef struct {
	uint8 selector_field; //4:0
	uint8 able100F; // 8
	uint8 able100H; // 7
	uint8 able10F; // 6
	uint8 able10H; // 5
} tcphy_mr4_reg_t; // use for mr4 & mr5

typedef struct {
	uint8 lp_np_able; // 3
	uint8 lp_autoneg_able; // 0
} tcphy_mr6_reg_t; // use for mr6

typedef struct {
	uint8  slicer_err_thd; // 15:11
	uint16 err_over_cnt;   // 10:0
	uint16 err_over_cnt_prev; // 10:0
} tcphy_l0r25_reg_t;

typedef struct {
	uint8 lch_sig_detect;
	uint8 lch_rx_linkpulse;
	uint8 lch_linkup_100;
	uint8 lch_linkup_10;
	uint8 lch_linkup_mdi;
	uint8 lch_linkup_mdix;
	uint8 lch_descr_lock;
	uint8 mdix_status;
	uint8 tx_amp_save;
	uint8 final_duplex;
	uint8 final_speed;
	uint8 final_link;	
} tcphy_l0r28_reg_t;

tcphy_mr0_reg_t mr0;
tcphy_mr4_reg_t mr4; 
tcphy_mr4_reg_t mr5;
tcphy_mr6_reg_t mr6;
tcphy_l0r28_reg_t mr28; // L0R28

// for multiple phy support
tcphy_mr1_reg_t Nmr1[TCPHY_PORTNUM];
tcphy_l0r25_reg_t Nmr25[TCPHY_PORTNUM];

// type for register settings
typedef struct cfg_data_s{
	uint32 reg_num;
	uint32 val;
}cfg_data_t;
typedef struct cfg_cL45data_s
{
    uint32 	dev_num;
    uint32 	reg_num;
    uint32 	val;
}cfg_cl45data_t;
typedef struct cfg_trdata_s
{
    char 	reg_typ[10];
    uint32 	reg_num;
    uint32 	val;
}cfg_trdata_t;

uint8 	allGeAnaCalWait(uint32 delay, uint8 phyaddr);  		
void	GECal_Rext(uint8 phyaddr, uint32 delay);			
uint32 	mtEMiiRegRead(uint32 port_num, uint32 dev_num, uint32 reg_num);
uint32 	tcMiiExtStationRead_CL22(uint32 port_num, uint32 dev_num, uint32 reg_num);


uint32 
tcPhyReadReg(uint8 port_num, uint8 reg_num)
{
	uint32 val, val_r31;
	uint32 phyAddr = mac_p->enetPhyAddr + port_num;

	// page to L0
	val_r31 = tcMiiStationRead(phyAddr, 31); 
	tcMiiStationWrite(phyAddr, 31, 0x8000); 	
	// read reg
	val = tcMiiStationRead(phyAddr, reg_num); 
	// restore page if necessary
	if (val_r31 != 0x8000) {
		tcMiiStationWrite(phyAddr, 31, val_r31);
	}

	// update variables
	switch(reg_num){
		
	case 0:
		mr0.main_reset = (val>>15)&0x00000001;
		mr0.force_speed    = (val>>13)&0x00000001;
		mr0.autoneg_enable = (val>>12)&0x00000001;
		mr0.powerdown      = (val>>11)&0x00000001;
		mr0.force_duplex   = (val>>8)&0x00000001;
		break;
		
	case 1:
		Nmr1[port_num].value = val;
		Nmr1[port_num].link_status_prev = Nmr1[port_num].link_status;
		Nmr1[port_num].link_status = (val>>2)&0x00000001;
		break;

	case 4:
		mr4.able100F = (val>>8)&0x0001; 	
		mr4.able100H = (val>>7)&0x0001; 	
		mr4.able10F = (val>>6)&0x0001; 	
		mr4.able10H = (val>>5)&0x0001; 	
		mr4.selector_field = (val)&0x001f;
		break;

	case 5:
		mr5.able100F = (val>>8)&0x0001; 	
		mr5.able100H = (val>>7)&0x0001; 	
		mr5.able10F = (val>>6)&0x0001; 	
		mr5.able10H = (val>>5)&0x0001; 	
		mr5.selector_field = (val)&0x001f;
		break;
		
	case 6:		
		mr6.lp_np_able = (val>>3)&0x0001;
		mr6.lp_autoneg_able = (val)&0x0001;
		break;
		
	case 25:		
		Nmr25[port_num].err_over_cnt_prev = Nmr25[port_num].err_over_cnt;
		Nmr25[port_num].err_over_cnt = (val & 0x0000007ff);
		break;
		
	case 28:
		mr28.lch_sig_detect  = (val>>15)&0x0001;
		mr28.lch_rx_linkpulse= (val>>14)&0x0001;
		mr28.lch_linkup_100  = (val>>13)&0x0001;
		mr28.lch_linkup_10   = (val>>12)&0x0001;
		mr28.lch_linkup_mdi  = (val>>11)&0x0001; // after LEM
		mr28.lch_linkup_mdix = (val>>10)&0x0001; // after LEM
		mr28.lch_descr_lock  = (val>>9)&0x0001; // after LEM
		mr28.mdix_status  = (val>>5)&0x0001; /* {0:mdi,1:mdix} */	
		mr28.tx_amp_save  = (val>>3)&0x0003; /* 0:100%, 1:90%, 2:80%, 3:70% */
		mr28.final_duplex = (val>>2)&0x0001; /* {0:half-duplex, 1:full-duplex} */
		mr28.final_speed  = (val>>1)&0x0001; /* {0:10, 1:100} */
		mr28.final_link   = (val)&0x0001; /* {0:linkdown, 1:linkup} */		
		break;
		
	default:
		break;
	}	
	return (val);
}


// read Global Reg
uint32
tcPhyReadGReg(uint8 port_num,uint8 page_num,uint8 reg_num){
	uint32 val, val_r31;
	uint32 phyAddr = mac_p->enetPhyAddr + port_num;
	uint32 pageAddr = (page_num<<12);

	// page to L0
	val_r31 = tcMiiStationRead(phyAddr, 31); 
	tcMiiStationWrite(phyAddr, 31, pageAddr); 	
	val = tcMiiStationRead(phyAddr, reg_num); 
	// restore page if necessary
	if (val_r31 != pageAddr) {
		tcMiiStationWrite(phyAddr, 31, val_r31);
	}
	
	return val;
}

void
tcPhyWriteGReg(uint8 port_num,uint8 page_num,uint8 reg_num,uint32 reg_data){
	uint32 val_r31;
	uint32 phyAddr = mac_p->enetPhyAddr + port_num;
	uint32 pageAddr = (page_num<<12);

	// page to L0
	val_r31 = tcMiiStationRead(phyAddr, 31); 
	tcMiiStationWrite(phyAddr, 31, pageAddr); 	
	tcMiiStationWrite(phyAddr, reg_num, reg_data); 
	// restore page if necessary
	if (val_r31 != pageAddr) {
		tcMiiStationWrite(phyAddr, 31, val_r31);
	}
}

// write Local Reg
void
tcPhyWriteLReg(uint8 port_num,uint8 page_num,uint8 reg_num,uint32 reg_data){
	uint32 val_r31;
	uint32 phyAddr = mac_p->enetPhyAddr + port_num;
	uint32 pageAddr = (page_num<<12)+0x8000;

	// page to L0
	val_r31 = tcMiiStationRead(phyAddr, 31); 
	tcMiiStationWrite(phyAddr, 31, pageAddr); 	
	tcMiiStationWrite(phyAddr, reg_num, reg_data); 
	// restore page if necessary
	if (val_r31 != pageAddr) {
		tcMiiStationWrite(phyAddr, 31, val_r31);
	}
}

bool 
tc2104mcInitialRegFlag(void)
{
	bool initial_value_flag=1; // default: correct initial value
	uint8 pn;
	uint16 rval;

	// check G3R16
	rval = tcPhyReadGReg(0,3,16);
	if (rval != 0x0018){
		initial_value_flag=0; // wrong initial value
	}
	// check L0R25
	for(pn=0;pn<TCPHY_PORTNUM;pn++){
		rval = tcPhyReadReg(pn,25);
		if ((rval&0xf000)!=0xc000){
			initial_value_flag=0; // wrong initial value
		}
	}
	return (initial_value_flag);
}

bool getTcPhyInitialRegFlag(void){
	if (!sw_patch_flag)
		return 0;

	switch (tcPhyVer) {
		case tcPhyVer_2104mc: // 2206
			return (tc2104mcInitialRegFlag());
			break;
	}
	return 0;
}

/************************************************************************
*
*            Variables and Functions for 7530 GE PHY (ID=0x9412)
*
**************************************************************************/
#define MTPHY_PORTNUM 5


#define MT7530_PORTNUM 					5
#define MT7530_PHY_INIT_LDATA_LEN 		1
#define MT7530_PHY_INIT_PERPDATA_LEN 	0
#define MT7530_PHY_INIT_SET_NUM 		1

#define MT7530_PHY_INIT_CL45_GDATA_LEN 		1	// allen_20150206	
#define MT7530_PHY_INIT_CL45_LDATA_LEN 		1	// allen_20150206
#define MT7530_PHY_INIT_CL45_PERPDATA_LEN 	0
#define MT7530_PHY_INIT_CL45_SET_NUM 		1

#define MT7530_PHY_INIT_TR_LDATA_LEN 		0
#define MT7530_PHY_INIT_TR_PERPDATA_LEN 	0
#define MT7530_PHY_INIT_TR_SET_NUM 			1

typedef struct mt7530_cfg_data_s
{
	char name[10];
	cfg_data_t ldata[MT7530_PHY_INIT_LDATA_LEN];
	cfg_data_t perpdata[MT7530_PHY_INIT_PERPDATA_LEN]; //per port register setting
}mt7530_cfg_data_t;

typedef struct mt7530_cfg_cl45data_s
{
	cfg_cl45data_t gdata[MT7530_PHY_INIT_CL45_GDATA_LEN];
	cfg_cl45data_t ldata[MT7530_PHY_INIT_CL45_LDATA_LEN];
	cfg_cl45data_t perpdata[MT7530_PHY_INIT_CL45_PERPDATA_LEN]; //per port register setting
}mt7530_cfg_cl45data_t;

typedef struct mt7530_cfg_trdata_s
{
	cfg_trdata_t ldata[MT7530_PHY_INIT_TR_LDATA_LEN];
	cfg_trdata_t perpdata[MT7530_PHY_INIT_PERPDATA_LEN]; //per port register setting
}mt7530_cfg_trdata_t;

const uint32 mt7530_page_sel_addr = 31;

mt7530_cfg_data_t mt7530_cfg[MT7530_PHY_INIT_SET_NUM] = 
{
    {
    	{"E4.0"},
	    //local data	
     	{ 	
			{0x09, 0x1e00}	// master mode
		},
    	//per-port data 
     	{ 
	 	}
    }, 
};

mt7530_cfg_cl45data_t mt7530_cfg_cl45[MT7530_PHY_INIT_CL45_SET_NUM]=
{
    {	//globle data
     	{
			{0x1f,0x0417,0x7775}  // allen_20150206
		},
     	//local data
     	{
			{0x1e,0x00a6,0x0350}   
		},
	    //per-port data
     	{
	 	}
    }, 
};

mt7530_cfg_trdata_t mt7530_cfg_tr[MT7530_PHY_INIT_TR_SET_NUM]=
{
    {	//local data
     	{ 
		},
     	//per-port data
		{
        }
    }, 
};

void tcMiiExtStationWrite_CL22(uint32 port_num,uint32 dev_num,uint32 reg_num,uint32 reg_data)
{
    const uint16 MMD_Control_register=0xD;
    const uint16 MMD_addr_data_register=0xE;
  	const uint16 page_reg=31;
	tcMiiStationWrite(port_num, page_reg, 0x00); //switch to main page
	tcMiiStationWrite(port_num, MMD_Control_register, (0<<14)+dev_num);
	tcMiiStationWrite(port_num, MMD_addr_data_register, reg_num);
	tcMiiStationWrite(port_num, MMD_Control_register, (1<<14)+dev_num);
	tcMiiStationWrite(port_num, MMD_addr_data_register, reg_data);
}

void mtEMiiRegWrite(uint32 port_num, uint32 dev_num, uint32 reg_num, uint32 reg_data)
{
	tcMiiExtStationWrite_CL22( port_num, dev_num, reg_num, reg_data);
}

void mt7530GePhyCfgLoad(uint8 idx, uint8 doPerPort_flag, uint8 port_num)
{
	int pn, i;
	uint16 phyAddr;
	current_idx = idx;

    // global registers
    phyAddr = port_num; //ephy_addr_base;
	for(pn=0; pn<MTPHY_PORTNUM; pn++)
	{
		phyAddr = pn; //ephy_addr_base + pn;

		for( i=0; i<MT7530_PHY_INIT_CL45_GDATA_LEN; i++ )
		{         
			mtEMiiRegWrite(phyAddr, mt7530_cfg_cl45[current_idx].gdata[i].dev_num, mt7530_cfg_cl45[current_idx].gdata[i].reg_num, mt7530_cfg_cl45[current_idx].gdata[i].val);
		}
		
		for( i=0; i<MT7530_PHY_INIT_CL45_LDATA_LEN; i++ )
		{         
			mtEMiiRegWrite(phyAddr, mt7530_cfg_cl45[current_idx].ldata[i].dev_num, mt7530_cfg_cl45[current_idx].ldata[i].reg_num, mt7530_cfg_cl45[current_idx].ldata[i].val);
		}
		
		for( i=0; i<MT7530_PHY_INIT_LDATA_LEN; i++ )
		{         
			tcMiiStationWrite(phyAddr, mt7530_cfg[current_idx].ldata[i].reg_num, mt7530_cfg[current_idx].ldata[i].val);
		}
		
		for( i=0; i<MT7530_PHY_INIT_TR_LDATA_LEN; i++ )
		{         
			mtPhyMiiWrite_TrDbg( phyAddr, mt7530_cfg_tr[current_idx].ldata[i].reg_typ, mt7530_cfg_tr[current_idx].ldata[i].reg_num, mt7530_cfg_tr[current_idx].ldata[i].val,0);
		}
	}  
}

/************************************************************************
*
*            Variables and Functions for 7512 FE PHY (ID=0x9412, R15(0xf)[13]=0)
*
**************************************************************************/
#define MT7512FE_PHY_INIT_GDATA_LEN 	5	
#define MT7512FE_PHY_INIT_LDATA_LEN 	9
#define MT7512FE_PHY_INIT_PERPDATA_LEN 	1
#define MT7512FE_PHY_INIT_SET_NUM 		1
#define MT7512FE_PORTNUM 				4

typedef struct mt7512FE_cfg_data_s
{
    char 		name[10];
    cfg_data_t 	gdata[MT7512FE_PHY_INIT_GDATA_LEN];
    cfg_data_t 	ldata[MT7512FE_PHY_INIT_LDATA_LEN];
	cfg_data_t 	perpdata[MT7512FE_PHY_INIT_PERPDATA_LEN]; 	//per port register setting
}mt7512FE_cfg_data_t;

const uint32 mt7512FE_page_sel_addr = 31;
const uint16 mt7512FE_G0R22_Tx10AmpSave_ON  = 0x0264;
const uint16 mt7512FE_G0R22_Tx10AmpSave_OFF = 0x006F;
const uint16 mt7512FE_L2R16[MT7512FE_PORTNUM] = {0x0606, 0x0606, 0x0606, 0x0606, 0x0606};  		
const uint16 mt7512FE_L2R17_A1[MT7512FE_PORTNUM] = {0x0003, 0x0003, 0x0003, 0x0003, 0x0003};  	
const uint16 mt7512FE_L3R17_A1[MT7512FE_PORTNUM] = {0x0000, 0x0000, 0x0000, 0x0000, 0x0000}; 	

mt7512FE_cfg_data_t mt7512FE_cfg[MT7512FE_PHY_INIT_SET_NUM] = 
{
	{ 
		{"Fe_1.0"},
		{ 	/* global */
        	{31,0x4000},           	// g4  // allen_20130926 : [5:4] reduce cr_a_tx_cmd_cld for 10Base 3th harmonic 
        	//{16,0x00cc}
        	{26,0x8044},			// g4r26: tx bias (0x8444)
        	{31,0x5000},
        	{21,0x00ea},			// g5r21: z2p_ovs
        	{27,0x02f0}				// g5r27: z2n_ovs
     	},    
     	{  							// local
     		{31,0x8000}, 			// l0
     		{30,0x0000},			// l0r30: tx pwd
     		{31,0x9000},			// l1
     		{22,0xf000},			// l1r22: for intel82579 EEE IOT
     		{31,0xa000},			// l2
     		{22,0x4444},			// l2r22: ps_dri (0x3444)
     		{28,0x7c44},			// l2r28: ps_op (0x6c44)
			{31,0xb000},  			// l3
			{17,0x0000}   			//disable EEE
     	},
     	{ 
     		{31,0x8000}
	 	} 
    }, 
};

static void mt7512FECfgTx10AmpSave(void)
{
    uint16 phyAddr = ephy_addr_base;

    if(cfg_Tx10AmpSave_flag==1)
	{ 	// enable
        tcPhyWriteGReg(phyAddr,0,22,mt7512FE_G0R22_Tx10AmpSave_ON);
    }
    else 
	{ 	// disable
        tcPhyWriteGReg(phyAddr,0,22,mt7512FE_G0R22_Tx10AmpSave_OFF);
    }
}

static void mt7512FELRCfgLoad(uint8 idx, uint8 port_num)
{
    int i;
	current_idx = idx;

	if(ephy_addr_base == 0)   // allen_20140911
		ephy_addr_base += 8;
	
	for(i=0; i<MT7512FE_PHY_INIT_LDATA_LEN; i++ )
	{         
    	tcMiiStationWrite((port_num+ephy_addr_base), mt7512FE_cfg[current_idx].ldata[i].reg_num, mt7512FE_cfg[current_idx].ldata[i].val );
    }   
}

// 1. loading default register setting in EPhyInit()
// 2. changing register setting
void mt7512FECfgLoad(uint8 idx, uint8 doPerPort_flag, uint8 phyaddr)
{
    int pn,i;
    uint16 port_num;
	current_idx = idx;
	
	if(ephy_addr_base == 0)
		ephy_addr_base += 8; 

	//printf("phyaddr = %d \r\n", phyaddr);

    for(i=0; i<MT7512FE_PHY_INIT_GDATA_LEN; i++ )
	{         
        tcMiiStationWrite(phyaddr, mt7512FE_cfg[current_idx].gdata[i].reg_num, mt7512FE_cfg[current_idx].gdata[i].val );
    }
    
    mt7512FECfgTx10AmpSave();
	
    // local registers 
    if(doPerPort_flag == DO_PER_PORT) 
	{
		port_num = phyaddr - ephy_addr_base;  
		mt7512FELRCfgLoad(current_idx, port_num);
    }
	else 
	{
		 for(pn=0; pn<MT7512FE_PORTNUM; pn++)
		 {
		 	mt7512FELRCfgLoad(current_idx, pn);
		}
	}
}


/************************************************************************
*
*            Variables and Functions for 7512 GE PHY (ID=0x9451)
*
**************************************************************************/

#define MT7512Ge_PORTNUM 				1
#define MT7512Ge_PHY_INIT_LDATA_LEN 	1
#define MT7512Ge_PHY_INIT_PERPDATA_LEN 	0
#define MT7512Ge_INIT_SET_NUM 			1
#define MT7512Ge_PHY_INIT_SET_NUM 		1

#define MT7512Ge_PHY_INIT_CL45_GDATA_LEN 	9 	
#define MT7512Ge_PHY_INIT_CL45_LDATA_LEN 	9 	 
#define MT7512Ge_PHY_INIT_CL45_PERPDATA_LEN 0
#define MT7512Ge_PHY_INIT_CL45_SET_NUM 		1

#define MT7512Ge_PHY_INIT_TR_LDATA_LEN 		0
#define MT7512Ge_PHY_INIT_TR_PERPDATA_LEN 	0
#define MT7512Ge_PHY_INIT_TR_SET_NUM 		1

typedef struct mt7512Ge_cfg_data_s
{
	char 		name[10];
	cfg_data_t 	ldata[MT7512Ge_PHY_INIT_LDATA_LEN];
	cfg_data_t 	perpdata[MT7512Ge_PHY_INIT_PERPDATA_LEN]; //per port register setting	
}mt7512Ge_cfg_data_t;

typedef struct mt7512Ge_cfg_cl45data_s
{
	cfg_cl45data_t gdata[MT7512Ge_PHY_INIT_CL45_GDATA_LEN];
	cfg_cl45data_t ldata[MT7512Ge_PHY_INIT_CL45_LDATA_LEN];
	cfg_cl45data_t perpdata[MT7512Ge_PHY_INIT_CL45_PERPDATA_LEN]; //per port register setting	
}mt7512Ge_cfg_cl45data_t;

typedef struct mt7512Ge_cfg_trdata_s
{
	cfg_trdata_t ldata[MT7512Ge_PHY_INIT_TR_LDATA_LEN];
	cfg_trdata_t perpdata[MT7512Ge_PHY_INIT_PERPDATA_LEN]; //per port register setting	
}mt7512Ge_cfg_trdata_t;

const uint32 mt7512Ge_page_sel_addr = 31;

mt7512Ge_cfg_data_t mt7512Ge_cfg[MT7512Ge_PHY_INIT_SET_NUM]=
{
	{
		{"Ge_1.0"}, 
    	//local data	
     	{ 	
       		{0x09, 0x1e00}
     	},
    	//per-port data 
	    { 
	 	}
    },
};

mt7512Ge_cfg_cl45data_t mt7512Ge_cfg_cl45[MT7512Ge_PHY_INIT_CL45_SET_NUM]=
{
    {	//globle data
     	{
			{0x1f, 0x0272, 0x3cff},		// ps_op
			{0x1f, 0x0269, 0x4344},		// ps_dri
			{0x1f, 0x0044, 0x0030},		// pair delay (digital)

			{0x1f, 0x027c, 0x0808},		// 10 base Tx
			{0x1f, 0x027b, 0x1177},		// disable 10 base-Te
			
			{0x1f, 0x0417, 0x7775},
			{0x1f, 0x0024, 0xc007},		// gphy led
			{0x1f, 0x0025, 0x003f},		// gphy led
			{0x1f, 0x0021, 0x800a}		// gphy led
     	},
     	
     	//local data
     	{
			{0x1e,0x0002,0x01b4},	// z2p_ovs
			{0x1e,0x0005,0x0080},	// p2z_ovs
			{0x1e,0x0008,0x03b7},	// z2n_ovs
			{0x1e,0x000b,0x0002},	// n2z_ovs
			
			{0x1e,0x0014,0x0440},	// pair delay (afe)

			{0x1e,0x0012,0x2c10},	// 10 base Tx
			{0x1e,0x0017,0x0410},	// 10 base Tx
			
			{0x1e,0x00a6,0x0350},
			{0x07,0x003c,0x0000}
	     },
     	//per-port data
     	{
	 	}
    }, 
};

mt7512Ge_cfg_trdata_t mt7512Ge_cfg_tr[MT7512Ge_PHY_INIT_TR_SET_NUM]=
{
    {	//local data
     	{ 
     	},

     	//per-port data
     	{ 
	 	}
    }, 
};

void mt7512GePhyCfgLoad(uint8 idx)
{
	int pn, i;
	uint16 phyAddr, phyAddr_base = 12;
	current_idx = idx;

	eco_rev = VPint(0xbfb00064);

    // global registers
    for( pn=0; pn<tcPhyPortNum; pn++)
	{
		phyAddr = phyAddr_base + pn;
	
		//printf("\n 7512Ge, phyaddr= (%d,%d) \n", phyAddr_base, phyAddr);
		for( i=0; i<MT7512Ge_PHY_INIT_CL45_GDATA_LEN; i++ )
		{         
			mtEMiiRegWrite(phyAddr, mt7512Ge_cfg_cl45[current_idx].gdata[i].dev_num, mt7512Ge_cfg_cl45[current_idx].gdata[i].reg_num, mt7512Ge_cfg_cl45[current_idx].gdata[i].val);
		}
		for( i=0; i<MT7512Ge_PHY_INIT_CL45_LDATA_LEN; i++ )
		{         
			mtEMiiRegWrite(phyAddr, mt7512Ge_cfg_cl45[current_idx].ldata[i].dev_num, mt7512Ge_cfg_cl45[current_idx].ldata[i].reg_num, mt7512Ge_cfg_cl45[current_idx].ldata[i].val);
		}
		for( i=0; i<MT7512Ge_PHY_INIT_LDATA_LEN; i++ )
		{         
			tcMiiStationWrite(phyAddr, mt7512Ge_cfg[current_idx].ldata[i].reg_num, mt7512Ge_cfg[current_idx].ldata[i].val);
		}
		for( i=0; i<MT7512Ge_PHY_INIT_TR_LDATA_LEN; i++ )
		{         
			mtPhyMiiWrite_TrDbg( phyAddr, mt7512Ge_cfg_tr[current_idx].ldata[i].reg_typ, mt7512Ge_cfg_tr[current_idx].ldata[i].reg_num, mt7512Ge_cfg_tr[current_idx].ldata[i].val,0);
		}
	}
}


// tcPhy initial: reset, load default register setting, restat AN
int tcPhyInit(macAdapter_t * mac_p){
	int i=0;
	uint16 start_addr= mac_p->enetPhyAddr;

	for( i=start_addr; i<start_addr+tcPhyPortNum; i++ ){			
		tcMiiStationWrite(i, PHY_CONTROL_REG, PHY_RESET );
	}

	switch (tcPhyVer) {
		case mtPhyVer_7530: 		
            mt7530GePhyCfgLoad(0, DO_5_PORT, 0);
		    break;
        case tcPhyVer_7512Fe:
		case tcPhyVer_7522Fe:
            mt7512FECfgLoad(0, DO_4_PORT, 8);
            break;
            
        case tcPhyVer_7512Ge:
		case tcPhyVer_7522Ge: 		
            mt7512FECfgLoad(0, DO_4_PORT, 8);
			mt7512GePhyCfgLoad(0);   
            break;
            
	    default:
	    	printf("Not found TC Phy\r\n");
	    	break;
	    	
	}

	for( i=start_addr; i<start_addr+tcPhyPortNum; i++ ){			
		tcMiiStationWrite(i, PHY_CONTROL_REG, MIIDR_AUTO_NEGOTIATE );
	}

	for(i=0;i<TCPHY_PORTNUM;i++){
		tcphy_link_state[i]=ST_LINK_DOWN;
	}

	return 0;
}
int tcPhyPortNumGet(void){
	if(tcPhyVer != 99)
		return tcPhyPortNum;
	else
		return 0;
}


// set tcPhyFlag & tcPhyVer
int tcPhyVerLookUp(macAdapter_t *mac_p){
	uint32 rval;
    uint16 r15_val;
	tcPhyFlag = 1;
	rval = miiStationRead(mac_p->enetPhyAddr, 3); // phy revision id
	rval &= 0xffff;
	r15_val = miiStationRead(mac_p->enetPhyAddr, 0xF);
	if (rval==0x9400){
		tcPhyVer=tcPhyVer_2031; 		
		tcPhyPortNum=1;
		TCPHYDISP1("TC2031, ");
	} else if (rval==0x9401){
		tcPhyVer=tcPhyVer_2101mb;
		tcPhyPortNum=1;
		//printf("TC2101MB, ");
	} else if (rval==0x9402){
		tcPhyVer=tcPhyVer_2104mc;
		tcPhyPortNum=4;
		//printf("TC2104MC, ");
	} else if (rval==0x9403){
		tcPhyVer=tcPhyVer_2104sd;
		tcPhyPortNum=4;
		//printf("TC2104SD, ");
	} else if (rval==0x9404){
		tcPhyVer=tcPhyVer_2101me;
		tcPhyPortNum=1;
		//printf("TC2101ME, ");
	 } else if (rval==0x9405){
        tcPhyVer=tcPhyVer_2102me;
        tcPhyPortNum=2;
	 } else if (rval==0x9406){
        tcPhyVer=tcPhyVer_2104me;
        tcPhyPortNum=4;
	} else if (rval==0x9407){
        tcPhyVer=tcPhyVer_2101mf;
        tcPhyPortNum=1;
   } else if (rval==0x940b){	
        tcPhyVer=tcPhyVer_2105mj;
        tcPhyPortNum=5;
	} else if (rval==0x940F){	 
		 tcPhyVer=tcPhyVer_mt7510FE;
		 tcPhyPortNum=4;
		 printf("tcPhyVer_mt7510FE \n ");	 
	 } else if (rval==0x9412){ 
        if(r15_val & 0x2000)
        {
            tcPhyVer=mtPhyVer_7530;
            tcPhyPortNum=1;
        }else{
            tcPhyVer=tcPhyVer_7512Fe;
            tcPhyPortNum=4;
	    } 
	 } else if (rval==0x9421){ 
		 tcPhyVer=mtPhyVer_7510Ge;
		 tcPhyPortNum=1;
    }else if (rval==0x9451){
        tcPhyVer=tcPhyVer_7512Ge;
		 tcPhyPortNum=1;
	}
	else if(rval == 0x9414)  
	{
        tcPhyVer = tcPhyVer_7522Fe;		
        tcPhyPortNum = 3;
    } 
    else if(rval == 0x9471)  
	{
        tcPhyVer = tcPhyVer_7522Ge;		
        tcPhyPortNum = 1;
	}
	return 0;
}


uint32
macPhyLookUp(
	macAdapter_t *mac_p,
	uint32 companyId
)
{
	uint32 i;
	uint32 phyTypeSupport;
        phyTypeSupport = sizeof(phyDeviceList) / sizeof(phyDeviceList_t);
	for ( i = 0; i < phyTypeSupport; i++ ) {
		if ( companyId == phyDeviceList[i].companyId ) {
			mac_p->enetPhyId = i;
			if( companyId == 0x03a2 ){
				tcPhyVerLookUp(mac_p);
			} 
			return 1;
		}
	}
	return 0;
}

int32
macSearchPhyAddr(
	macAdapter_t *mac_p
)
{
	uint32 miiReg = 0;
	int cnt=10000;
	do{
		if(miiStationRead(0 , 2)!=0xffff){
			break;
		}
		cnt--;
	}while(cnt>0);
	{
		miiReg = miiStationRead(mac_p->enetPhyAddr , PHY_ID_REG1);
		if (miiReg == 0)
			miiReg = miiStationRead(mac_p->enetPhyAddr, PHY_ID_REG2);
		if ( macPhyLookUp(mac_p, miiReg) ) {
			return 0;
		}
	}
	return -1;
}
/*_____________________________________________________________________________
**      function name: filedSwicDefVal
**      descriptions:
**       We need to refilled  the TC2206 switch ic register default value, it be
**         overwrited by hardware trap.
**            
**      parameters:
**         none
**             
**      global:
**         None
**             
**      return:
**             None
**	     
**      call:
**   	     None
**      		
**      revision:
**      1. Here 2009/7/30
**      2. Here 2010/1/30
**____________________________________________________________________________
*/
void
filedSwicDefVal(void){
 	int i=0;
	uint16 ver=0;
	uint16 val=0;
	/*Get the TC2206 switch IC model id*/
	ver=(uint16)(1<<(tcMiiStationRead(31, 30)));

	for(i=0; tc2206_swic_reg[i].phyaddr!=0; i++){
		if(ver & tc2206_swic_reg[i].model_id){
			tcMiiStationWrite(tc2206_swic_reg[i].phyaddr, tc2206_swic_reg[i].regaddr, tc2206_swic_reg[i].value);
		}
	}
}/*end filedSwicDefVal*/

/*_____________________________________________________________________________
**      function name: macSetUpPhy
**      descriptions:
**           Detect the switch vender id and setup related mii register,such as
 **          cpu port's link up status.
**             
**      parameters:
**            mac_p: 
**         
**      global:
**            None
**             
**      return:
**            0
**	     
**      call:
**      	miiStationRead
**      	miiStationWrite
**      	
**      revision:
**      2009/05/08
**____________________________________________________________________________
*/
int
macSetUpPhy
(
	macAdapter_t *mac_p
)
{
	/* detect if RTL8305 or RTL8306 Switch*/	
	if ((miiStationRead(4, 2) == 0x1c) && (miiStationRead(4, 3) == 0xc852)) {
		if ((miiStationRead(6, 2) == 0x1c) && (miiStationRead(6, 3) == 0xc852) ){
    			/*swicVendor = SWIC_RTL8306SD;Let CPU Port Link up*/
    			miiStationWrite(5, 22,miiStationRead(5,22)|0x8000);
    			miiStationWrite(6, 22,miiStationRead(6,22)|0x8000);    		
        	}
	}
	tcPhyInit(mac_p);
	/*Switch Model Number Register (SMNR), Addr.: 16'h02FE*/
	if (tcMiiStationRead(31, 31) == 0x2206){
	    /*TC2206 switch IC and to filled the switch default value*/
	    swicVendor=SWIC_TC2206;
	    filedSwicDefVal();
	}


  return 0;
}/*end macSetUpPhy*/
/*_____________________________________________________________________________
**      function name: getTC2206InitRegFlag
**      descriptions:
**       System will accroding the esd_chk_reg to check the switch status is handing or not
**            
**      parameters:
**         none
**             
**      global:
**         esd_chk_reg
**             
**      return:
**             0: Switch IC is happend handing problem.
**             1: Work fine.
**	     
**      call:
**   	     tcMiiStationRead
**      		
**      revision:
**      1. Here 2009/8/6
**____________________________________________________________________________
*/
int 
getTC2206InitRegFlag(void){
	int i=0;
	int j=0;
	uint16 ver = 0;
	ver = (uint16)(1<<(tcMiiStationRead(31,30)));
	for(i=0; esd_chk_reg[i].phyaddr!=0; i++){
		for(j=0; tc2206_swic_reg[j].phyaddr!=0; j++){
			/*Check the switch ic register is reset to default*/
			if((esd_chk_reg[i].model_id & ver)
			&&(tc2206_swic_reg[j].model_id)
			&&(esd_chk_reg[i].phyaddr == tc2206_swic_reg[j].phyaddr) 
			&&(esd_chk_reg[i].regaddr == tc2206_swic_reg[j].regaddr)){
				if(tcMiiStationRead(esd_chk_reg[i].phyaddr, esd_chk_reg[i].regaddr)!=tc2206_swic_reg[j].value){
					return 0;
				}
			}
		}
	
	}
	return 1;
}/*end getTC2206InitRegFlag*/

/*_____________________________________________________________________________
**      function name: esdSwPatch
**      descriptions:
**       During the ESD test that is due to switch/phy handing on at TC2206 switch ic, so we need to use 
**	   hardware reset to wake up the hardware.So system need to reload the switch/phy register settings.
**	   
**      parameters:
**         none
**             
**      global:
**         None
**             
**      return:
**             None
**	     
**      call:
**   	     None
**      		
**      revision:
**      1. Here 2009/7/30
**____________________________________________________________________________
*/
void 
esdSwPatch(void){
	uint8 ether_gpio=*(uint8 *)(FLASH_ETHER_GPIO);
	if((swicVendor==SWIC_TC2206) && esd_sw_patch && (miiStationRead(31,30)==0x0)){
		if((!getTC2206InitRegFlag()) ||(!getTcPhyInitialRegFlag())){
			printf("\r\nESD detected!!\r\n");
			/* ----- Hardware reset Ehernet phy chip, this address is defined by h/w engineer ----- */
			VPint(CR_GPIO_CTRL) |= (1<<(ether_gpio<<1));/*Example: gpio 9=(1<<(9*2))*/
			VPint(CR_GPIO_ODRAIN) |= (1<<ether_gpio);

			VPint(CR_GPIO_DATA) &= ~(1<<ether_gpio);
			pause(100);

			VPint(CR_GPIO_DATA) |= (1<<ether_gpio);
			pause(100);

			/*Reload the switch default value for hardware trap*/
			filedSwicDefVal();
			/*Reload the tce phy default value for afe settings and power issue*/
			tcPhyInit(mac_p);
		}
	}	
}/*end filedSwicDefVal*/

/*=======================*/
void macPhyReset(void)
{
	{
		miiStationWrite(8,0x1f,0x8000);
		miiStationWrite(8,0x1e,0x2000);
		miiStationWrite(9,0x1f,0x8000);
		miiStationWrite(9,0x1e,0x2000);
		miiStationWrite(10,0x1f,0x8000);
		miiStationWrite(10,0x1e,0x2000);
		miiStationWrite(11,0x1f,0x8000);
		miiStationWrite(11,0x1e,0x2000);
	}

	return;
}
void macPhyReset_tc3162l2(void){
	eth_gpioReset();
	miiStationWrite(mac_p->enetPhyAddr, PHY_CONTROL_REG, PHY_RESET );
	miiStationWrite(mac_p->enetPhyAddr, PHY_CONTROL_REG, MIIDR_AUTO_NEGOTIATE );
}

void macPhyReset_tc3262g(void){
	eth_gpioReset();
	miiStationWrite(mac_p->enetPhyAddr, PHY_CONTROL_REG, PHY_RESET );
	pause(100);
	miiStationWrite(mac_p->enetPhyAddr, PHY_CONTROL_REG, PHY_RESTART_AN | PHY_AN_ENABLE);
}

void macPhyReset_mac(void){
	miiStationWrite(mac_p->enetPhyAddr, PHY_CONTROL_REG, PHY_RESET );
	miiStationWrite(mac_p->enetPhyAddr, PHY_CONTROL_REG, MIIDR_AUTO_NEGOTIATE );
}

/*=====from 7512_eth.c to tcswitch.c*/
#define ANACAL_INIT			0x01
#define ANACAL_ERROR		0xFD
#define ANACAL_SATURATION	0xFE
#define	ANACAL_FINISH		0xFF
#define read_reg_word(reg) 				VPint(reg)

void	GECal_Rext(uint8 phyaddr, uint32 delay)									
{
	uint8	rg_zcal_ctrl, all_ana_cal_status;
	uint16	ad_cal_comp_out_init;
	uint16	dev1e_e0_ana_cal_r5; 
	int 	calibration_polarity;
	uint16	reg_temp;
	uint8	cnt=0;

	// *** Iext/Rext Cal start ********************************************************************************
	all_ana_cal_status = ANACAL_INIT;
	// analog calibration enable, Rext calibration enable
	mtEMiiRegWrite(phyaddr, 0x1e, 0x00db, 0x1110);								// 1e_db[12]:rg_cal_ckinv, [8]:rg_ana_calen, [4]:rg_rext_calen, [0]:rg_zcalen_a
	mtEMiiRegWrite(phyaddr, 0x1e, 0x00dc, 0x0000);								// 1e_dc[0]:rg_txvos_calen
	mtEMiiRegWrite(phyaddr, 0x1e, 0x00e1, 0x0000);								// 1e_e1[4]:rg_cal_refsel(0:1.2V)

	rg_zcal_ctrl = 0x20;  														// start with 0 dB
	dev1e_e0_ana_cal_r5 = mtEMiiRegRead(phyaddr, 0x1e, 0x00e0);
	mtEMiiRegWrite(phyaddr, 0x1e, 0x00e0, (rg_zcal_ctrl));						// 1e_e0[5:0]:rg_zcal_ctrl

	all_ana_cal_status = allGeAnaCalWait(delay, phyaddr); 						// delay 20 usec
	if(all_ana_cal_status == 0)
	{
		all_ana_cal_status = ANACAL_ERROR;	
		printf(" GE Rext AnaCal ERROR!   \r\n");
	}
	
	ad_cal_comp_out_init = (mtEMiiRegRead(phyaddr, 0x1e, 0x017a)>>8) & 0x1;		// 1e_17a[8]:ad_cal_comp_out	
	if(ad_cal_comp_out_init == 1)
	{
		calibration_polarity = -1;
	}
	else 																		// ad_cal_comp_out_init == 0
	{
		calibration_polarity = 1;
	}
	
	cnt = 0;
	while(all_ana_cal_status < ANACAL_ERROR)
	{
		cnt ++;
		rg_zcal_ctrl += calibration_polarity;
		mtEMiiRegWrite(phyaddr, 0x1e, 0x00e0, (rg_zcal_ctrl));
		all_ana_cal_status = allGeAnaCalWait(delay, phyaddr); 					// delay 20 usec

		if(all_ana_cal_status == 0)
		{
			all_ana_cal_status = ANACAL_ERROR;	
			printf("  GE Rext AnaCal ERROR!   \r\n");
		}
		else if(((mtEMiiRegRead(phyaddr, 0x1e, 0x017a)>>8)&0x1) != ad_cal_comp_out_init) 
		{
			all_ana_cal_status = ANACAL_FINISH;	
			//printf("  GE Rext AnaCal Done! (%d)(0x%x)  \r\n", cnt, rg_zcal_ctrl);
		}
		else
		{
			if((rg_zcal_ctrl == 0x3F)||(rg_zcal_ctrl == 0x00))	
			{
				all_ana_cal_status = ANACAL_SATURATION;  						// need to FT(IC fail?)
				printf(" GE Rext AnaCal Saturation!  \r\n");
				rg_zcal_ctrl = 0x20;  											// 0 dB
			}
			//else
				//printf(" GE Rxet cal (%d)(%d)(%d)(0x%x)  \r\n", cnt, ad_cal_comp_out_init, ((mtEMiiRegRead(phyaddr, 0x1e, 0x017a)>>8)&0x1), mtEMiiRegRead(phyaddr, 0x1e, 0xe0));
		}
	}

	if(all_ana_cal_status == ANACAL_ERROR)
	{	
		rg_zcal_ctrl = 0x20;  													// 0 dB
		mtEMiiRegWrite(phyaddr, 0x1e, 0x00e0, (dev1e_e0_ana_cal_r5 | rg_zcal_ctrl));
	}
	else
	{
		mtEMiiRegWrite(phyaddr, 0x1e, 0x00e0, (dev1e_e0_ana_cal_r5 | rg_zcal_ctrl));
		mtEMiiRegWrite(phyaddr, 0x1e, 0x00e0, ((rg_zcal_ctrl<<8)|rg_zcal_ctrl));
		//****  1f_115[2:0] = rg_zcal_ctrl[5:3]  								// Mog review
		mtEMiiRegWrite(phyaddr, 0x1f, 0x0115, ((rg_zcal_ctrl & 0x3f)>>3));
		printf("  GE Rext AnaCal Done! (%d)(0x%x)  \r\n", cnt, rg_zcal_ctrl);
	}
	mtEMiiRegWrite(phyaddr, 0x1e, 0x00db, 0x0000);
	// *** Iext/Rext Cal end ***
}
uint8 allGeAnaCalWait(uint32 delay, uint8 phyaddr) 
{
	uint8 all_ana_cal_status;	
	
	mtEMiiRegWrite(phyaddr, 0x1e, 0x017c, 0x0001);								// da_calin_flag pull high
	
	//udelay(delay);
	pause(1);
	all_ana_cal_status = mtEMiiRegRead(phyaddr, 0x1e, 0x017b) & 0x1;
	mtEMiiRegWrite(phyaddr, 0x1e, 0x017c, 0x0000);								// da_calin_flag pull low

	return all_ana_cal_status;
}

uint32 mtEMiiRegRead(uint32 port_num, uint32 dev_num, uint32 reg_num)
{
	return(tcMiiExtStationRead_CL22( port_num, dev_num, reg_num));
}

uint32 tcMiiExtStationRead_CL22(uint32 port_num, uint32 dev_num, uint32 reg_num)
{
    const uint16 MMD_Control_register=0xD;  	// 0xd=13
    const uint16 MMD_addr_data_register=0xE;	// 0xe=14	
  	const uint16 page_reg=31;
	uint32 value=0;
	miiStationWrite(port_num, page_reg, 0x00); //switch to main page
	miiStationWrite(port_num, MMD_Control_register, (0<<14)+dev_num);
	miiStationWrite(port_num, MMD_addr_data_register, reg_num);
	miiStationWrite(port_num, MMD_Control_register, (1<<14)+dev_num);
	value = miiStationRead(port_num, MMD_addr_data_register);
	return(value); 	
}



int eth_phy_init(macAdapter_t *mac_p){
	uint32	efuse_bit31_0_val, efuse_bit63_32_val;
	int i, err = 0;
	if (isFPGA){
		printf("fpga mode, phy init quit\n");
	}
	for ( mac_p->enetPhyAddr = 0; mac_p->enetPhyAddr < 32; mac_p->enetPhyAddr++ ) {
		if(!macSearchPhyAddr(mac_p)){
			macPhyReset();
			
			if (macSetUpPhy(mac_p)){
				err = -1;
			}
	
		}
	}
	macPhyReset();
	return err;
}

int eth_phy_init_7520(macAdapter_t *mac_p){

	int err =0;
		for ( mac_p->enetPhyAddr = 0; mac_p->enetPhyAddr < 32; mac_p->enetPhyAddr++ ) {
			if(!macSearchPhyAddr(mac_p)){
				if (macSetUpPhy(mac_p)){
					err = -1;
				}
				mac_p->enetPhyAddr += (tcPhyPortNumGet()-1);
					break;
			}
			
		}
		mac_p->enetPhyAddr = 0;

	return err;
}

int eth_phy_init_tc3162l2(macAdapter_t *mac_p){
	int err = 0;
	macPhyReset_tc3162l2();
	pause(100);

	macSearchPhyAddr(mac_p);
	/* ----- Initialize the SEEQ 80225 phy chip(enet phy) ----- */
	if (macSetUpPhy(mac_p)){
		err = -1;
	}

	return err;
}

int eth_phy_init_tc3262g(macAdapter_t *mac_p){
	int err = 0;
	macPhyReset_tc3262g();
	pause(100);

	macSearchPhyAddr(mac_p);
	printf("Search PHY addr and found PHY addr=%d\n", mac_p->enetPhyAddr);
	if (macSetUpPhy(mac_p)){
		err = -1;
	}
	return err;
}

void eth_phy_init_mac(void){

	macPhyReset_mac();
}

